import copy
from dataclasses import dataclass
from datetime import datetime, timedelta
from dateutil import parser
from enum import auto, Enum, unique
import json


@dataclass
class Subscription:
    topic: str
    snapshot: timedelta


@dataclass
class Unsubscription:
    topic: str


@unique
class Operation(Enum):
    ADD = "ADD"
    REMOVE = "REMOVE"


@unique
class MessageType(Enum):
    INTEL = auto()
    SIGHTING = auto()


@unique
class IntelType(Enum):
    IPSRC = auto()
    IPDST = auto()
    IPSRC_PORT = auto()
    IPDST_PORT = auto()
    EMAILSRC = auto()
    EMAILDST = auto()
    TARGETEMAIL = auto()
    EMAILATTACHMENT = auto()
    FILENAME = auto()
    HOSTNAME = auto()
    DOMAIN = auto()
    DOMAIN_IP = auto()
    URL = auto()
    URI = auto()
    USERAGENT = auto()
    MD5 = auto()
    MALWARESAMPLE = auto()
    FILENAME_MD5 = auto()
    SHA1 = auto()
    FILENAME_SHA1 = auto()
    SHA256 = auto()
    FILENAME_SHA256 = auto()
    X509FINGERPRINTSHA1 = auto()
    PDB = auto()
    AUTHENTIHASH = auto()
    SSDEEP = auto()
    IMPHASH = auto()
    PEHASH = auto()
    IMPFUZZY = auto()
    SHA224 = auto()
    SHA384 = auto()
    SHA512 = auto()
    SHA512_224 = auto()
    SHA512_256 = auto()
    TLSH = auto()
    CDHASH = auto()
    FILENAME_AUTHENTIHASH = auto()
    FILENAME_SSDEEP = auto()
    FILENAME_IMPHASH = auto()
    FILENAME_PEHASH = auto()
    FILENAME_IMPFUZZY = auto()
    FILENAME_SHA224 = auto()
    FILENAME_SHA384 = auto()
    FILENAME_SHA512 = auto()
    FILENAME_SHA512_224 = auto()
    FILENAME_SHA512_256 = auto()
    FILENAME_TLSH = auto()


class IntelData(dict):
    """Threat Bus intel data is a dictionary with at least two keys: 'indicator' and 'intel_type'.
    The 'indicator' is a tuple of strings, defining the IoC(s)
    The 'intel_type' is a threatbus.data.IntelType
    """

    def __init__(self, indicator, intel_type: IntelType, *args, **kw):
        super(IntelData, self).__init__(*args, **kw)
        assert indicator, "Intel indicator must be set"
        assert (
            isinstance(indicator, tuple)
            and all(map(lambda e: isinstance(e, str), indicator))
            or isinstance(indicator, str)
        ), "Intel indicator must either be a string or tuple of strings"
        assert intel_type, "Intel type must be set"
        assert isinstance(
            intel_type, IntelType
        ), "Intel type must be of type threatbus.data.IntelType"
        self["indicator"] = indicator if isinstance(indicator, tuple) else (indicator,)
        self["intel_type"] = intel_type


@dataclass()
class Intel:
    ts: datetime
    id: str
    data: IntelData
    operation: Operation


@dataclass()
class Sighting:
    ts: datetime
    intel: str
    context: dict


@dataclass()
class SnapshotEnvelope:
    """
    SnapshotEnvelopes are used to wrap intel items or sightings in response to
    SnapshotRequests. The envelope carries additional information, such as the
    MessageType and the unique ID of a snapshot.
    """

    snapshot_type: MessageType
    snapshot_id: str
    body: Intel or Sighting


@dataclass
class SnapshotRequest:
    """
    Threat Bus creates SnapshotRequests when apps specifically as for a snapshot
    during subscription. SnapshotRequests are provisioned via the backbones.
    A request consists of the requested MessageType (either INTEL or SIGHTING),
    a unique ID, and the snapshot time delta (e.g., 30 days).
    """

    snapshot_type: MessageType
    snapshot_id: str
    snapshot: timedelta


class IntelEncoder(json.JSONEncoder):
    """
    Encodes Intel objects to JSON strings
    """

    def default(self, intel: Intel):
        if type(intel) is not Intel:
            # let the base class default method raise the TypeError
            return json.JSONEncoder.default(self, intel)
        data = copy.deepcopy(intel.data)
        data["indicator"] = list(intel.data["indicator"])
        data["intel_type"] = int(intel.data["intel_type"].value)
        return {
            "ts": str(intel.ts),
            "id": str(intel.id),
            "data": data,
            "operation": str(intel.operation.value),
        }


class IntelDecoder(json.JSONDecoder):
    """
    Decodes JSON strings to Intel objects
    """

    def __init__(self, *args, **kwargs):
        json.JSONDecoder.__init__(self, object_hook=self.decode_hook, *args, **kwargs)

    def decode_hook(self, dct: dict):
        if "intel_type" in dct and "indicator" in dct:
            # parse IntelData
            intel_type = IntelType(int(dct.pop("intel_type")))
            indicator = tuple(dct.pop("indicator"))
            return IntelData(indicator, intel_type, **dct)
        elif "ts" in dct and "id" in dct and "data" in dct and "operation" in dct:
            # parse Intel
            return Intel(
                parser.parse(dct["ts"]),
                dct["id"],
                dct["data"],
                Operation(dct["operation"]),
            )
        return dct


class SightingEncoder(json.JSONEncoder):
    """
    Encodes Sighting objects to JSON strings
    """

    def default(self, sighting: Sighting):
        if type(sighting) is not Sighting:
            # let the base class default method raise the TypeError
            return json.JSONEncoder.default(self, sighting)
        return {
            "ts": str(sighting.ts),
            "intel": sighting.intel,
            "context": sighting.context,
        }


class SightingDecoder(json.JSONDecoder):
    """
    Decodes JSON strings to Sighting objects
    """

    def __init__(self, *args, **kwargs):
        json.JSONDecoder.__init__(self, object_hook=self.decode_hook, *args, **kwargs)

    def decode_hook(self, dct: dict):
        if "ts" in dct and "intel" in dct and "context" in dct:
            return Sighting(parser.parse(dct["ts"]), dct["intel"], dct["context"])
        return dct


class SnapshotRequestEncoder(json.JSONEncoder):
    """
    Encodes SnapshotRequest objects to JSON strings
    """

    def default(self, req: SnapshotRequest):
        if type(req) is not SnapshotRequest:
            # let the base class default method raise the TypeError
            return json.JSONEncoder.default(self, req)
        return {
            "snapshot_type": int(req.snapshot_type.value),
            "snapshot_id": str(req.snapshot_id),
            "snapshot": int(req.snapshot.total_seconds()),
        }


class SnapshotRequestDecoder(json.JSONDecoder):
    """
    Decodes JSON strings to SnapshotRequest objects
    """

    def __init__(self, *args, **kwargs):
        json.JSONDecoder.__init__(self, object_hook=self.decode_hook, *args, **kwargs)

    def decode_hook(self, dct: dict):
        if "snapshot_type" in dct and "snapshot_id" in dct and "snapshot" in dct:
            return SnapshotRequest(
                MessageType(int(dct["snapshot_type"])),
                dct["snapshot_id"],
                timedelta(seconds=dct["snapshot"]),
            )
        return dct


class SnapshotEnvelopeEncoder(json.JSONEncoder):
    """
    Encodes SnapshotEnvelope objects to JSON strings
    """

    def default(self, env: SnapshotEnvelope):
        if type(env) is not SnapshotEnvelope:
            # let the base class default method raise the TypeError
            return json.JSONEncoder.default(self, env)
        encoder = None
        if type(env.body) is Intel:
            encoder = IntelEncoder
        elif type(env.body) is Sighting:
            encoder = SightingEncoder
        else:
            # let the base class default method raise the TypeError
            return json.JSONEncoder.default(self, env)
        return {
            "snapshot_type": int(env.snapshot_type.value),
            "snapshot_id": str(env.snapshot_id),
            "body": encoder.default(self, env.body),
        }


class SnapshotEnvelopeDecoder(json.JSONDecoder):
    """
    Decodes JSON strings to SnapshotEnvelope objects
    """

    def __init__(self, *args, **kwargs):
        json.JSONDecoder.__init__(self, object_hook=self.decode_hook, *args, **kwargs)

    def decode_hook(self, dct: dict):
        if "snapshot_type" in dct and "snapshot_id" in dct and "body" in dct:
            snapshot_type = MessageType(int(dct["snapshot_type"]))
            return SnapshotEnvelope(
                snapshot_type,
                dct["snapshot_id"],
                dct["body"],
            )
        if (
            "intel_type" in dct
            and "indicator" in dct
            or "ts" in dct
            and "id" in dct
            and "data" in dct
            and "operation" in dct
        ):
            return IntelDecoder.decode_hook(self, dct)
        elif "ts" in dct and "intel" in dct and "context" in dct:
            return SightingDecoder.decode_hook(self, dct)
        return dct



import coloredlogs
import logging
import sys


def setup(config, name):
    fmt = "%(asctime)s %(levelname)-8s [%(name)s] %(message)s"
    colored_formatter = coloredlogs.ColoredFormatter(fmt)
    plain_formatter = logging.Formatter(fmt)
    logger = logging.getLogger(name)
    if config["file"]:
        fh = logging.FileHandler(config["filename"].get(str))
        fhLevel = logging.getLevelName(config["file_verbosity"].get(str).upper())
        logger.setLevel(fhLevel)
        fh.setLevel(fhLevel)
        fh.setFormatter(plain_formatter)
        logger.addHandler(fh)
    if config["console"]:
        ch = logging.StreamHandler()
        chLevel = logging.getLevelName(config["console_verbosity"].get(str).upper())
        ch.setLevel(chLevel)
        if logger.level > chLevel or logger.level == 0:
            logger.setLevel(chLevel)
        ch.setFormatter(colored_formatter)
        logger.addHandler(ch)

    class ShutdownHandler(logging.Handler):
        """Exit application with CRITICAL logs"""

        def emit(self, record):
            logging.shutdown()
            sys.exit(1)

    sh = ShutdownHandler(level=50)
    sh.setFormatter(colored_formatter)
    logger.addHandler(sh)
    return logger



import argparse
import confuse
import pluggy
from queue import Queue
from threatbus import appspecs, backbonespecs, logger
from threatbus.data import MessageType, SnapshotRequest, SnapshotEnvelope
from threading import Lock
from uuid import uuid4


class ThreatBus:
    def __init__(self, backbones, apps, logger, config):
        self.backbones = backbones
        self.apps = apps
        self.config = config
        self.logger = logger
        self.inq = Queue()  # fan-in everything, provisioned by backbone
        self.snapshot_q = Queue()
        self.lock = Lock()
        self.snapshots = dict()

    def handle_snapshots(self):
        """
        Waits to handle snapshot requests or envelopes. Forwards new requests to
        all implementing app plugins. Forwards envelopes to the requesting app
        or discards them accordingly.
        """
        while True:
            msg = self.snapshot_q.get(block=True)
            if type(msg) is SnapshotRequest:
                self.logger.debug(f"Received SnapshotRequest: {msg}")
                self.apps.snapshot(snapshot_request=msg, result_q=self.inq)
            elif type(msg) is SnapshotEnvelope:
                self.logger.debug(f"Received SnapshotEnvelope: {msg}")
                if msg.snapshot_id not in self.snapshots:
                    continue
                self.snapshots[msg.snapshot_id].put(msg.body)
            else:
                self.logger.warn(
                    f"Received message with unknown type on snapshot topic: {msg}"
                )
            self.snapshot_q.task_done()

    def request_snapshot(self, topic, dst_q, time_delta):
        """
        Create a new SnapshotRequest and push it to the inq, so that the
        backbones can provision it.
        @param topic The topic for which the snapshot is requested
        @param dst_q A queue that should be used to forward all snapshot
            data to
        @param time_delta A timedelta object to mark the snapshot size
        """
        # Threat Bus follows a hierarchical pub-sub structure. Subscriptions
        # to 'threatbus' must hence result in snapshots for both, intel
        # and sightings
        message_types = []
        prefix = "threatbus"
        if topic == prefix:
            message_types = [MessageType.INTEL, MessageType.SIGHTING]
        elif topic.endswith("sighting"):
            message_types.append(MessageType.SIGHTING)
        elif topic.endswith("intel"):
            message_types.append(MessageType.INTEL)
        for mt in message_types:
            self.logger.info(
                f"Requesting snapshot from all plugins for message type {mt.name} and time delta {time_delta}"
            )
            snapshot_id = str(uuid4())
            self.snapshots[snapshot_id] = dst_q  # store queue of requester
            req = SnapshotRequest(mt, snapshot_id, time_delta)

            self.inq.put(req)

    def subscribe(self, topic, q, time_delta=None):
        """
        Accepts a new subscription for a given topic and queue pointer.
        Forwards that subscription to all managed backbones.
        @param topic Subscribe to this topic
        @param q A queue object to forward all messages for the given topics
        @param time_delta A timedelta object to mark the snapshot size
        """
        assert isinstance(topic, str), "topic must be string"
        self.backbones.subscribe(topic=topic, q=q)
        if time_delta:
            self.request_snapshot(topic, q, time_delta)

    def unsubscribe(self, topic, q):
        """
        Removes subscription for a given topic and queue pointer from all
        managed backbones.
        """
        assert isinstance(topic, str), "topic must be string"
        self.backbones.unsubscribe(topic=topic, q=q)

    def run(self):
        self.logger.info("Starting plugins...")
        logging = self.config["logging"]
        self.apps.run(
            config=self.config["plugins"]["apps"],
            logging=logging,
            inq=self.inq,
            subscribe_callback=self.subscribe,
            unsubscribe_callback=self.unsubscribe,
        )
        self.backbones.run(
            config=self.config["plugins"]["backbones"], logging=logging, inq=self.inq
        )
        self.subscribe("threatbus/snapshotrequest", self.snapshot_q)
        self.subscribe("threatbus/snapshotenvelope", self.snapshot_q)
        self.handle_snapshots()


def validate_config(config):
    if config["logging"]["console"].get(bool):
        config["logging"]["console_verbosity"].get(str)
    if config["logging"]["file"].get(bool):
        config["logging"]["file_verbosity"].get(str)
        config["logging"]["filename"].get(str)


def start(config):
    try:
        validate_config(config)
    except Exception as e:
        raise ValueError("Invalid config: {}".format(str(e)))

    backbones = pluggy.PluginManager("threatbus.backbone")
    backbones.add_hookspecs(backbonespecs)
    backbones.load_setuptools_entrypoints("threatbus.backbone")

    apps = pluggy.PluginManager("threatbus.app")
    apps.add_hookspecs(appspecs)
    apps.load_setuptools_entrypoints("threatbus.app")

    tb_logger = logger.setup(config["logging"], "threatbus")
    configured_apps = set(config["plugins"]["apps"].keys())
    installed_apps = set(dict(apps.list_name_plugin()).keys())
    for unwanted_app in installed_apps - configured_apps:
        tb_logger.info(f"Disabling installed, but unconfigured app '{unwanted_app}'")
        apps.unregister(name=unwanted_app)
    configured_backbones = set(config["plugins"]["backbones"].keys())
    installed_backbones = set(dict(backbones.list_name_plugin()).keys())
    for unwanted_backbones in installed_backbones - configured_backbones:
        tb_logger.info(
            f"Disabling installed, but unconfigured backbones '{unwanted_backbones}'"
        )
        backbones.unregister(name=unwanted_backbones)

    bus = ThreatBus(backbones.hook, apps.hook, tb_logger, config)
    bus.run()


def main():
    config = confuse.Configuration("threatbus")
    parser = argparse.ArgumentParser()
    parser.add_argument("--config", "-c", help="path to a configuration file")
    args = parser.parse_args()
    config.set_args(args)
    if args.config:
        config.set_file(args.config)
    start(config)


if __name__ == "__main__":
    main()



import pluggy

hookspec = pluggy.HookspecMarker("threatbus.app")


@hookspec
def run(config, logging, inq, subscribe_callback, unsubscribe_callback):
    """
    Runs / starts a plugin spec with a configuration object
    @param config A configuration object for the app
    @param logging A configuration object for the logger
    @param inq A queue in which this plugin puts incoming messages
    @param subscribe_callback A callback to use by a plugin to notify about
        new subscriptions
    @param unsubscribe_callback A callback to use by a plugin to notify
        about subscription revocations
    """


@hookspec
def snapshot(snapshot_request, result_q):
    """
    Perform a snapshot, based on the given `snapshot_request`. Snapshots are
    collected up to the requested earliest date. Results of the type
    threatbus.data.SnapshotEnvelope are put to the result_q.
    @param snapshot_request @see threatbus.data.SnapshotRequest
    @param result_q Snapshot results are forwarded to this queue
    """



from datetime import datetime, timedelta
import unittest
import json

from data import (
    Intel,
    IntelData,
    IntelType,
    MessageType,
    Operation,
    Sighting,
    SnapshotEnvelope,
    SnapshotRequest,
    IntelEncoder,
    IntelDecoder,
    SightingEncoder,
    SightingDecoder,
    SnapshotRequestEncoder,
    SnapshotRequestDecoder,
    SnapshotEnvelopeEncoder,
    SnapshotEnvelopeDecoder,
)


class TestJsonConversions(unittest.TestCase):
    def setUp(self):
        self.ts = datetime.now().astimezone()
        self.intel_id = "intel-42"
        self.indicator = "6.6.6.6"
        self.intel_type = IntelType.IPSRC
        self.operation = Operation.ADD
        self.intel_data = IntelData(
            self.indicator, self.intel_type, foo=23, more_args="MORE ARGS"
        )
        self.intel = Intel(self.ts, self.intel_id, self.intel_data, self.operation)

        self.sighting_context = {
            "ts": "2017-03-03T23:56:09.652643840",
            "uid": "CMeLkt11aTqwgN4FI9",
            "id.orig_h": "172.31.130.19",
            "id.orig_p": 43872,
            "id.resp_h": "172.31.129.17",
            "id.resp_p": 20004,
            "proto": "tcp",
            "service": None,
            "duration": 0.025249,
            "orig_bytes": 311,
            "resp_bytes": 999,
            "conn_state": "SF",
            "local_orig": None,
            "local_resp": None,
            "missed_bytes": 0,
            "history": "ShADadFf",
            "orig_pkts": 9,
            "orig_ip_bytes": 787,
            "resp_pkts": 7,
            "resp_ip_bytes": 1371,
            "tunnel_parents": [],
            "alert": {
                "signature": "VAST-RETRO Generic IoC match for: 172.31.129.17",
                "category": "Potentially Bad Traffic",
                "action": "allowed",
            },
            "event_type": "alert",
            "_extra": {"vast-ioc": "172.31.129.17"},
            "source": "VAST",
        }
        self.sighting = Sighting(self.ts, self.intel_id, self.sighting_context)

        self.snapshot_id = "SNAPSHOT_UUID"
        self.snapshot = timedelta(days=42, hours=23, minutes=13, seconds=37)
        self.snapshot_request = SnapshotRequest(
            MessageType.SIGHTING, self.snapshot_id, self.snapshot
        )

        self.snapshot_envelope_intel = SnapshotEnvelope(
            MessageType.INTEL, self.snapshot_id, self.intel
        )
        self.snapshot_envelope_sighting = SnapshotEnvelope(
            MessageType.SIGHTING, self.snapshot_id, self.sighting
        )

    def test_valid_intel_encoding(self):
        encoded = json.dumps(self.intel, cls=IntelEncoder)
        # encoding cannot be compared as string, because order of the fileds in
        # the string representation is not fixed.
        # We cast the JSON back to a simple python dict and them compare values
        py_dict = json.loads(encoded)
        self.assertEqual(py_dict["ts"], str(self.ts))
        self.assertEqual(py_dict["id"], self.intel_id)
        self.assertEqual(
            py_dict["data"],
            {
                "foo": 23,
                "more_args": "MORE ARGS",
                "indicator": [self.indicator],
                "intel_type": self.intel_type.value,
            },
        )

    def test_valid_intel_decoding(self):
        encoded = f'{{"ts": "{self.ts}", "id": "intel-42", "data": {{"foo": 23, "more_args": "MORE ARGS", "indicator": ["{self.indicator}"], "intel_type": "{self.intel_type.value}"}}, "operation": "{self.operation.value}"}}'
        intel = json.loads(encoded, cls=IntelDecoder)
        self.assertEqual(intel, self.intel)

    def test_intel_encoding_roundtrip(self):
        encoded = json.dumps(self.intel, cls=IntelEncoder)
        read_back = json.loads(encoded, cls=IntelDecoder)
        self.assertEqual(read_back, self.intel)

    def test_valid_sighting_encoding(self):
        encoded = json.dumps(self.sighting, cls=SightingEncoder)
        # encoding cannot be compared as string, because order of the fileds in
        # the string representation is not fixed.
        # We cast the JSON back to a simple python dict and them compare values
        py_dict = json.loads(encoded)
        self.assertEqual(py_dict["ts"], str(self.ts))
        self.assertEqual(py_dict["intel"], self.intel_id)
        self.assertEqual(py_dict["context"], self.sighting_context)

    def test_valid_sighting_decoding(self):
        encoded = f'{{"ts": "{self.ts}", "intel": "{self.intel_id}", "context": {json.dumps(self.sighting_context)}}}'
        sighting = json.loads(encoded, cls=SightingDecoder)
        self.assertEqual(sighting, self.sighting)

    def test_sighting_encoding_roundtrip(self):
        encoded = json.dumps(self.sighting, cls=SightingEncoder)
        read_back = json.loads(encoded, cls=SightingDecoder)
        self.assertEqual(read_back, self.sighting)

    def test_valid_snapshot_request_encoding(self):
        encoded = json.dumps(self.snapshot_request, cls=SnapshotRequestEncoder)
        # encoding cannot be compared as string, because order of the fileds in
        # the string representation is not fixed.
        # We cast the JSON back to a simple python dict and them compare values
        py_dict = json.loads(encoded)
        self.assertEqual(py_dict["snapshot_type"], MessageType.SIGHTING.value)
        self.assertEqual(py_dict["snapshot_id"], self.snapshot_id)
        self.assertEqual(py_dict["snapshot"], self.snapshot.total_seconds())

    def test_valid_snapshot_request_decoding(self):
        encoded = f'{{"snapshot_type": {MessageType.SIGHTING.value}, "snapshot_id": "{self.snapshot_id}", "snapshot": {self.snapshot.total_seconds()}}}'
        read_back = json.loads(encoded, cls=SnapshotRequestDecoder)
        self.assertEqual(read_back, self.snapshot_request)

    def test_snapshot_request_encoding_roundtrip(self):
        encoded = json.dumps(self.snapshot_request, cls=SnapshotRequestEncoder)
        read_back = json.loads(encoded, cls=SnapshotRequestDecoder)
        self.assertEqual(read_back, self.snapshot_request)

    def test_valid_snapshot_envelope_encoding(self):
        encoded_envelope_intel = json.dumps(
            self.snapshot_envelope_intel, cls=SnapshotEnvelopeEncoder
        )
        encoded_envelope_sighting = json.dumps(
            self.snapshot_envelope_sighting, cls=SnapshotEnvelopeEncoder
        )
        # encoding cannot be compared as string, because order of the fileds in
        # the string representation is not fixed.
        # We cast the JSON back to a simple python dict and them compare values
        py_dict = json.loads(encoded_envelope_intel)
        self.assertEqual(py_dict["snapshot_type"], MessageType.INTEL.value)
        self.assertEqual(py_dict["snapshot_id"], self.snapshot_id)
        self.assertEqual(py_dict["body"]["ts"], str(self.ts))
        self.assertEqual(py_dict["body"]["id"], self.intel_id)
        self.assertEqual(py_dict["body"]["operation"], self.operation.value)
        self.assertEqual(py_dict["body"]["data"]["indicator"][0], self.indicator)
        self.assertEqual(py_dict["body"]["data"]["intel_type"], self.intel_type.value)

        py_dict = json.loads(encoded_envelope_sighting)
        self.assertEqual(py_dict["snapshot_type"], MessageType.SIGHTING.value)
        self.assertEqual(py_dict["snapshot_id"], self.snapshot_id)
        self.assertEqual(py_dict["body"]["ts"], str(self.ts))
        self.assertEqual(py_dict["body"]["intel"], self.intel_id)
        self.assertEqual(py_dict["body"]["context"], self.sighting_context)

    def test_valid_snapshot_envelope_decoding(self):
        encoded_envelope_intel = f'{{"snapshot_type": {MessageType.INTEL.value}, "snapshot_id": "{self.snapshot_id}", "body": {{"ts": "{self.ts}", "id": "{self.intel_id}", "data": {{"foo": 23, "more_args": "MORE ARGS", "indicator": ["{self.indicator}"], "intel_type": "{self.intel_type.value}"}}, "operation": "{self.operation.value}"}}}}'
        read_back = json.loads(encoded_envelope_intel, cls=SnapshotEnvelopeDecoder)
        self.assertEqual(read_back, self.snapshot_envelope_intel)

        encoded_envelope_sighting = f'{{"snapshot_type": {MessageType.SIGHTING.value}, "snapshot_id": "{self.snapshot_id}", "body": {{"ts": "{self.ts}", "intel": "{self.intel_id}", "context": {json.dumps(self.sighting_context)}}}}}'
        read_back = json.loads(encoded_envelope_sighting, cls=SnapshotEnvelopeDecoder)
        self.assertEqual(read_back, self.snapshot_envelope_sighting)

    def test_snapshot_envelope_encoding_roundtrip(self):
        encoded = json.dumps(self.snapshot_envelope_intel, cls=SnapshotEnvelopeEncoder)
        read_back = json.loads(encoded, cls=SnapshotEnvelopeDecoder)
        self.assertEqual(read_back, self.snapshot_envelope_intel)

        encoded = json.dumps(
            self.snapshot_envelope_sighting, cls=SnapshotEnvelopeEncoder
        )
        read_back = json.loads(encoded, cls=SnapshotEnvelopeDecoder)
        self.assertEqual(read_back, self.snapshot_envelope_sighting)
